/************************************************************************ *********
 * Proiectarea cu Microprocesoare
 * Laboratorul 1
 * Digital I/O
 *********************************************************************************/

#include <avr\io.h>
#include <avr\interrupt.h>
#define F_CPU 16000000
#include <util/delay.h>

#include <string.h>
#include <stdlib.h>

#include "lcd.h"

#define S_FACTOR 	3
#define NUM_ROWS 	8
#define NUM_COLS 	16
#define R_NUMBER	3
#define DEATH_EXP	3

/*
 * CONWAY   = 0
 * HIGHLIFE = 1
 */
int type = 1;

unsigned char custom[8][8] = {{0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0},			// 000
							{0x0,0x0,0x0,0x0,0x0,0x1f,0x1f,0x1f},			// 001
							{0x0,0x0,0x0,0x1f,0x1f,0x0,0x0,0x0},			// 010
							{0x0,0x0,0x0,0x1f,0x1f,0x1f,0x1f,0x1f},		// 011
							{0x1f,0x1f,0x1f,0x0,0x0,0x0,0x0,0x0},			// 100
							{0x1f,0x1f,0x1f,0x0,0x0,0x1f,0x1f,0x1f},		// 101
							{0x1f,0x1f,0x1f,0x1f,0x1f,0x0,0x0,0x0},		// 110
							{0x1f,0x1f,0x1f,0x1f,0x1f,0x1f,0x1f,0x1f}};	// 111
int changed_state = 0;

void init_board( char **board ){
	int i, j;
	if ( board == NULL )
		return;
	
	for ( i = 0 ; i < NUM_ROWS ; i++ ) {
		for ( j = 0 ; j < NUM_COLS ; j++ ) {
			if ( rand() % DEATH_EXP )
				board[i][j] = '0';
			else
				board[i][j] = '1';
		}
	}
}

int check_display_state(char **board, int i, int j){
	if (board[S_FACTOR * i][j] == '0') {
		if (board[S_FACTOR * i + 1][j] == '0') {
			if (board[S_FACTOR * i + 2][j] == '0')
				return 0;
			else
				return 1;
		} else {
			if (board[S_FACTOR * i + 2][j] == '0')
				return 2;
			else
				return 3;
		}
	} else {
		if (board[S_FACTOR * i + 1][j] == '0') {
			if (board[S_FACTOR * i + 2][j] == '0')
				return 4;
			else
				return 5;
		} else {
			if (board[S_FACTOR * i + 2][j] == '0')
				return 6;
			else
				return 7;
		}
	}
}

void display_board( char **board ){
	int j, s;
	if ( board == NULL )
		return;
	
	LCD_writeInstruction( LCD_INSTR_clearDisplay );
	
	for ( j = 0 ; j < NUM_COLS ; j++ ) {
		s = check_display_state(board, 0, j);
		LCD_writeData( s );
	}
	LCD_writeInstruction( LCD_INSTR_nextLine );
	for ( j = 0 ; j < NUM_COLS ; j++ ) {
		s = check_display_state(board, 1, j);
		LCD_writeData( s );
	}
}

int isAlive(char **board, int i, int j){
	if (i == -1)
		return isAlive(board, NUM_ROWS-1, j);
	if (i == NUM_ROWS)
		return isAlive(board, 0, j);
	if (j == -1)
		return isAlive(board, i, NUM_COLS-1);
	if (j == NUM_COLS)
		return isAlive(board, i, 0);
	return board[i][j] == '1';
}

int checkUpper(char **board, int i, int j){
	return isAlive(board, i+1, j-1) + isAlive(board, i+1, j) + isAlive(board, i+1, j+1);
}

int checkLower(char **board, int i, int j){
	return isAlive(board, i-1, j-1) + isAlive(board, i-1, j) + isAlive(board, i-1, j+1);
}

int checkRight(char **board, int i, int j){
	return isAlive(board, i, j+1);
}

int  checkLeft(char **board, int i, int j){
	return isAlive(board, i, j-1);
}

int checkState(char **board, int i, int j){
	return checkUpper(board, i, j) + checkLower(board, i, j) + checkLeft(board, i, j) + checkRight(board, i, j);
}

void transform(char **board){
	int i, j, neigh;
	
	for ( i = 0 ; i < NUM_ROWS ; i++ ) {
		for ( j = 0 ; j < NUM_COLS ; j++ ) {
			neigh = checkState(board, i, j);
			if (board[i][j] == '1') {
				if (neigh < 2 || neigh > 3) {
					board[i][j] = '0';
					changed_state = 1;
				}
			} else {
				if (type == 0) {					
					if (neigh == R_NUMBER) {
						board[i][j] = '1';
						changed_state = 1;
					}
				} else {
					if (neigh == R_NUMBER || neigh == 2 * R_NUMBER) {
						board[i][j] = '1';
						changed_state = 1;
					}
				}
			}
		}
	}
}

void init_custom_chars(){
	int i, j;
	
	for ( i = 0 ; i < 8 ; i++ ) {
		LCD_writeInstruction( 0x40 + ( i * 8 ) );
		for ( j = 0 ; j < 8 ; j++ )
			LCD_writeData( custom[i][j] );
		LCD_writeInstruction( 0x80 );
	}
}

int main(){
	char **board;
	int i;
	
	board = malloc( NUM_ROWS * sizeof( char* ) );
	for ( i = 0 ; i < NUM_ROWS ; i++ )
		board[i] = malloc( NUM_COLS * sizeof( char ) );

	/* Initializam LCD-ul si tabla de joc */
	LCD_init();
	init_board( board );
	init_custom_chars();

	/* Asteptam ca butonul de start sa fie apasat */
	while (1) {
		if (!(PIND & (1 << PD6)))
			break;
	}
	
	/* Iteram cate o generatie la fiecare 400ms */
	while (1) {
		display_board( board );
		transform( board );
		if ( changed_state ) {
			changed_state = 0;
			type = type ^ 1;
			init_board ( board );
		}
		_delay_ms(500);
	}
	
	/* Eliberam memoria alocata */
	for ( i = 0 ; i < NUM_ROWS ; i++ )
		free( board[i] );
	free( board );
	board = NULL;
	
	return 0;
}
